"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const zen_observable_1 = __importDefault(require("zen-observable"));
const observable_promise_1 = require("../observable-promise");
const implementation_1 = __importDefault(require("./implementation"));
const thread_1 = require("./thread");
exports.Thread = thread_1.Thread;
let nextPoolID = 1;
const hasSymbols = () => typeof Symbol === 'function';
const hasSymbol = (name) => hasSymbols() && Boolean(Symbol[name]);
function slugify(text) {
    return text.replace(/\W/g, " ").trim().replace(/\s+/g, "-");
}
var PoolEventType;
(function (PoolEventType) {
    PoolEventType["initialized"] = "initialized";
    PoolEventType["taskCompleted"] = "taskCompleted";
    PoolEventType["taskFailed"] = "taskFailed";
    PoolEventType["taskQueued"] = "taskQueued";
    PoolEventType["taskStart"] = "taskStart";
    PoolEventType["terminated"] = "terminated";
})(PoolEventType = exports.PoolEventType || (exports.PoolEventType = {}));
function createArray(size) {
    const array = [];
    for (let index = 0; index < size; index++) {
        array.push(index);
    }
    return array;
}
function findIdlingWorker(workers, maxConcurrency) {
    return workers.find(worker => worker.runningTasks.length < maxConcurrency);
}
function spawnWorkers(spawnWorker, count) {
    return createArray(count).map(() => ({
        init: spawnWorker(),
        runningTasks: []
    }));
}
function PoolConstructor(spawnWorker, optionsOrSize) {
    const options = typeof optionsOrSize === "number"
        ? { size: optionsOrSize }
        : optionsOrSize || {};
    const debug = debug_1.default(`threads:pool:${slugify(options.name || String(nextPoolID++))}`);
    const { concurrency = 1, size = implementation_1.default.defaultPoolSize } = options;
    let isClosing = false;
    let nextTaskID = 1;
    let runningTaskJobs = [];
    const taskQueue = [];
    const workers = spawnWorkers(spawnWorker, size);
    let eventSubject;
    const eventObservable = observable_promise_1.makeHot(new zen_observable_1.default(subscriber => {
        eventSubject = subscriber;
    }));
    Promise.all(workers.map(worker => worker.init)).then(() => eventSubject.next({
        type: PoolEventType.initialized,
        size: workers.length
    }), error => eventSubject.error(error));
    const scheduleWork = () => {
        debug(`Attempt de-queueing a task to run it...`);
        const availableWorker = findIdlingWorker(workers, concurrency);
        if (!availableWorker)
            return;
        const nextTask = taskQueue.shift();
        if (!nextTask)
            return;
        const workerID = workers.indexOf(availableWorker) + 1;
        debug(`Running task #${nextTask.id} on worker #${workerID}...`);
        eventSubject.next({
            type: PoolEventType.taskStart,
            taskID: nextTask.id,
            workerID
        });
        const run = () => __awaiter(this, void 0, void 0, function* () {
            try {
                const returnValue = yield nextTask.run(yield availableWorker.init);
                debug(`Task #${nextTask.id} completed successfully`);
                eventSubject.next({
                    type: PoolEventType.taskCompleted,
                    returnValue,
                    taskID: nextTask.id,
                    workerID
                });
            }
            catch (error) {
                debug(`Task #${nextTask.id} failed`);
                eventSubject.next({
                    type: PoolEventType.taskFailed,
                    taskID: nextTask.id,
                    error,
                    workerID
                });
                throw error;
            }
            finally {
                runningTaskJobs = runningTaskJobs.filter(someRunPromise => someRunPromise !== runPromise);
                if (!isClosing) {
                    scheduleWork();
                }
            }
        });
        const runPromise = run();
        runningTaskJobs.push(runPromise);
    };
    const pool = {
        events() {
            return eventObservable;
        },
        queue(taskFunction) {
            return __awaiter(this, void 0, void 0, function* () {
                if (isClosing) {
                    throw Error(`Cannot schedule pool tasks after terminate() has been called.`);
                }
                const task = {
                    id: nextTaskID++,
                    run: taskFunction
                };
                debug(`Queueing task #${task.id}...`);
                taskQueue.push(task);
                eventSubject.next({
                    type: PoolEventType.taskQueued,
                    taskID: task.id
                });
                return new Promise((resolve, reject) => {
                    const eventSubscription = pool.events().subscribe(event => {
                        if (event.type === PoolEventType.taskCompleted && event.taskID === task.id) {
                            eventSubscription.unsubscribe();
                            resolve(event.returnValue);
                        }
                        else if (event.type === PoolEventType.taskFailed && event.taskID === task.id) {
                            eventSubscription.unsubscribe();
                            reject(event.error);
                        }
                        else if (event.type === PoolEventType.terminated) {
                            eventSubscription.unsubscribe();
                            reject(Error("Pool has been terminated before task was run."));
                        }
                    });
                    try {
                        scheduleWork();
                    }
                    catch (error) {
                        eventSubscription.unsubscribe();
                        reject(error);
                    }
                });
            });
        },
        terminate(force) {
            return __awaiter(this, void 0, void 0, function* () {
                isClosing = true;
                if (!force) {
                    yield Promise.all(runningTaskJobs);
                }
                eventSubject.next({
                    type: PoolEventType.terminated,
                    remainingQueue: [...taskQueue]
                });
                eventSubject.complete();
                yield Promise.all(workers.map((worker) => __awaiter(this, void 0, void 0, function* () { return thread_1.Thread.terminate(yield worker.init); })));
            });
        }
    };
    if (hasSymbols() && hasSymbol("toStringTag")) {
        pool[Symbol.toStringTag] = () => `[object Pool]`;
    }
    return pool;
}
PoolConstructor.EventType = PoolEventType;
exports.Pool = PoolConstructor;
